/* Copyright 2022 Sensaru GmbH */

#ifndef LIBC1_NET_SRC_IQUEUE_H_
#define LIBC1_NET_SRC_IQUEUE_H_

#include <vector>
#include <cstdint>
#include <memory>
#include <mutex>
#include <thread>
#include <condition_variable>
#include <atomic>
#include <string>

namespace C1Net {

class IQueueEntry {
 public:
  IQueueEntry() = default;
  virtual ~IQueueEntry() = default;

  int64_t time = 0;
};

/**
 * This class implements a queue after the producer-consumer paradigm. It can manage one or more queues. Your class needs to be derived from @c IQueue to use it.
 */
class IQueue {
 public:
  /**
   * Constructor.
   *
   * @param queue_count The number of queues to initialize. The maximum value is 1'000'000.
   * @param buffer_size The maximum number of items allowed to be queued. The maximum size is 2'000'000'000.
   */
  IQueue(uint32_t queue_count, uint32_t buffer_size);
  virtual ~IQueue();

  /**
   * Starts the threads of a queue.
   *
   * @param index The index of the queue to start. The number of queues is defined by @c queueCount in the constructor.
   * @param wait_when_full When set to @c true, @c Enqueue() waits until the queue is empty enough to queue the provided item. This takes precedence over the argument @c waitWhenFull of @c Enqueue().
   * @param processing_thread_count The number of processing threads to start.
   */
  void StartQueue(int32_t index, bool wait_when_full, uint32_t processing_thread_count);

  /**
   * Starts the threads of a queue using a thread count that can be increased later.
   *
   * @param index The index of the queue to start. The number of queues is defined by @c queueCount in the constructor.
   * @param wait_when_full When set to @c true, @c Enqueue() waits until the queue is empty enough to queue the provided item. This takes precedence over the argument @c waitWhenFull of @c Enqueue().
   * @param initial_processing_thread_count The number of processing threads to start with the call of this method.
   * @param max_processing_thread_count The number maximum number of processing threads that can be started.
   */
  void StartQueue(int32_t index, bool wait_when_full, uint32_t initial_processing_thread_count, uint32_t max_processing_thread_count);

  /**
   * Stops the threads of a queue previously started with @c StartQueue().
   *
   * @param index The index of the queue to stop.
   */
  void StopQueue(int32_t index);

  /**
   * Checks if the specified queue has been started.
   */
  bool QueueIsStarted(int32_t index);

  /**
   * Start an additional thread.
   *
   * @param index The index of the queue to add the new thread.
   * @return Returns true when the thread could be started.
   */
  bool AddThread(int32_t index);

  /**
   * Enqueues an item.
   *
   * @param index The index of the queue to Enqueue the item into.
   * @param entry The item to queue.
   * @param wait_when_full When set to @c true the method waits until the queue is empty enough to queue the item. When @c waitWhenFull is set to @c true in @c StartQueue(), this argument is ignored.
   * @return Returns @c true if the item was successfully queued and @c false otherwise.
   */
  bool Enqueue(int32_t index, std::shared_ptr<IQueueEntry> &entry, bool wait_when_full = false);

  /**
   * This method is called by the processing threads for each item to Process. It needs to be implemented by the derived class.
   *
   * @param index The index of the queue.
   * @param entry The queued item to Process.
   */
  virtual void ProcessQueueEntry(int32_t index, std::shared_ptr<IQueueEntry> &entry) = 0;

  /**
   * Checks if a queue is empty.
   *
   * @param index The index of the queue to check.
   * @return Returns @c true if the queue is empty.
   */
  bool QueueEmpty(int32_t index);

  /**
   * Returns the number of processing threads.
   *
   * @return The number of processing threads for this queue.
   */
  uint32_t ProcessingThreadCount(int32_t index);

  /**
   * Returns the maximum number of processing threads.
   *
   * @return The number maximum number of processing threads for this queue.
   */
  uint32_t MaxProcessingThreadCount(int32_t index);

  /**
   * Returns the number of items queued in a queue.
   *
   * @param index The index of the queue to check.
   * @return Return the number of queued items.
   */
  int32_t QueueSize(int32_t index);

  double ThreadLoad(int32_t index);
  double MaxThreadLoad(int32_t index);
  double MaxThreadLoad1m(int32_t index);
  double MaxThreadLoad10m(int32_t index);
  double MaxThreadLoad1h(int32_t index);
  int64_t MaxWait(int32_t index);
  int64_t MaxWait1m(int32_t index);
  int64_t MaxWait10m(int32_t index);
  int64_t MaxWait1h(int32_t index);

  virtual void QueueLog(uint32_t log_level, const std::string &message) = 0;
 private:
  int32_t queue_count_ = 2;
  std::unique_ptr<std::atomic_bool[]> stop_processing_thread_;
  std::mutex add_thread_mutex_;
  int32_t buffer_size_ = 10000;
  std::vector<int32_t> buffer_head_;
  std::vector<int32_t> buffer_tail_;
  std::vector<int32_t> buffer_count_;
  std::vector<bool> wait_when_full_;
  std::vector<std::vector<std::shared_ptr<IQueueEntry>>> buffer_;
  std::unique_ptr<std::mutex[]> queue_mutex_ = nullptr;
  std::vector<std::vector<std::shared_ptr<std::thread>>> processing_thread_;
  std::unique_ptr<std::condition_variable[]> produce_condition_variable_ = nullptr;
  std::unique_ptr<std::condition_variable[]> processing_condition_variable_ = nullptr;

  std::unique_ptr<std::atomic<uint32_t>[]> threads_in_use_;

  std::atomic<uint32_t> dropped_entries_{0};
  std::atomic<int64_t> last_queue_full_error_{0};

  std::unique_ptr<std::atomic<double>[]> max_thread_load_1m_current_;
  std::unique_ptr<std::atomic<double>[]> max_thread_load_1m_;
  std::unique_ptr<std::atomic<int64_t>[]> max_wait_1m_current_;
  std::unique_ptr<std::atomic<int64_t>[]> max_wait_1m_;
  std::unique_ptr<std::atomic<int64_t>[]> last_1m_cycle_;

  std::unique_ptr<std::atomic<double>[]> max_thread_load_10m_current_;
  std::unique_ptr<std::atomic<double>[]> max_thread_load_10m_;
  std::unique_ptr<std::atomic<int64_t>[]> max_wait_10m_current_;
  std::unique_ptr<std::atomic<int64_t>[]> max_wait_10m_;
  std::unique_ptr<std::atomic<int64_t>[]> last_10m_cycle_;

  std::unique_ptr<std::atomic<double>[]> max_thread_load_1h_current_;
  std::unique_ptr<std::atomic<double>[]> max_thread_load_1h_;
  std::unique_ptr<std::atomic<int64_t>[]> max_wait_1h_current_;
  std::unique_ptr<std::atomic<int64_t>[]> max_wait_1h_;
  std::unique_ptr<std::atomic<int64_t>[]> last_1h_cycle_;

  std::unique_ptr<std::atomic<double>[]> max_thread_load_;
  std::unique_ptr<std::atomic<int64_t>[]> max_wait_;

  void Process(int32_t index);
};

}
#endif
