/* Copyright 2022 Sensaru GmbH */

#ifndef LIBC1_NET_SRC_TCPCLIENT_HPP_
#define LIBC1_NET_SRC_TCPCLIENT_HPP_

#include "Socket.h"
#include "TlsCommon.h"
#include "TcpCommon.h"

#include <vector>
#include <functional>
#include <mutex>
#include <atomic>
#include <thread>

namespace C1Net {

/**
 * Class to easily create a TCP client.
 *
 * TCP Client Example Code
 * =======================
 *
 * This example creates a simple TCP client that connects to https://www.google.com. The client sends an HTTP GET
 * request every 10 seconds and prints all incoming packets to standard output.
 *
 * Save the example below in `main.cpp` and compile with:
 *
 *     g++ -o main -std=c++17 main.cpp -lc1-net -lgcrypt -lgnutls
 *
 * Start the example:
 *
 *     ./main
 *
 * Example code:
 *
 *     #include <c1-net/TcpClient.h>
 *     #include <iostream>
 *
 *     void Log(uint32_t log_level, const std::string &message) {
 *       std::cout << message << std::endl;
 *     }
 *
 *     void PacketReceived(const C1Net::TcpPacket &packet) {
 *       std::cout << std::string(packet.begin(), packet.end()) << std::endl;
 *     }
 *
 *     int main() {
 *     C1Net::TcpClientInfo client_info;
 *       client_info.host = "www.google.com";
 *       client_info.port = 443;
 *       client_info.tls = true;
 *
 *       client_info.log_callback = std::bind(&Log, std::placeholders::_1, std::placeholders::_2);
 *       client_info.packet_received_callback = std::bind(&PacketReceived, std::placeholders::_1);
 *
 *       C1Net::TcpClient client(client_info);
 *       client.Start();
 *
 *       while (!client.Connected()) {
 *         std::this_thread::sleep_for(std::chrono::milliseconds(1));
 *       }
 *
 *       std::string request("GET / HTTP/1.1\r\nHost: www.google.com\r\nConnection: keep-alive\r\n\r\n");
 *       while (true) {
 *         client.Send({request.begin(), request.end()});
 *         std::this_thread::sleep_for(std::chrono::milliseconds(10000));
 *       }
 *     }
 */
class TcpClient {
 public:
  struct TcpClientInfo {
    std::string host;
    uint16_t port = 0;
    bool tls = false;
    bool verify_certificate = true;
    std::string ca_file;
    std::string ca_data;
    std::string client_cert_file;
    std::string client_cert_data;
    std::string client_key_file;
    std::string client_key_data;
    bool verify_custom_hostname = false;
    std::string custom_hostname;

    std::size_t connection_retries = 3;
    std::size_t read_timeout = 15000;
    std::size_t write_timeout = 15000;

    std::function<void(uint32_t log_level, const std::string &message)> log_callback;
    std::function<void(const TcpPacket &packet)> packet_received_callback;
  };

  explicit TcpClient(const TcpClientInfo &tcp_client_info);
  ~TcpClient();

  bool Connected();

  void Start();
  void Stop();
  void WaitForClientStopped();

  void Send(const TcpPacket &packet);
 private:
  TcpClientInfo tcp_client_info_;

  std::mutex socket_mutex_;
  PSocket socket_ = std::make_shared<Socket>(-1);

  //{{{ TLS
  bool tls_init_failed_ = false;
  PTlsSession tls_session_;
  std::mutex certificate_credentials_mutex_;

  PCertificateInfo certificates_;

  /**
   * Contains a copy of the current credentials in case the credentials are replaced while the connection is open.
   */
  std::shared_ptr<CertificateCredentials> current_client_certificate_credentials_;

  /**
   * Stores the certificate credentials so that they can be replaced at any time.
   */
  std::shared_ptr<CertificateCredentials> certificate_credentials_;
  //}}}

  std::atomic_bool connecting_{true};
  std::atomic_bool stop_client_{false};
  std::thread listen_thread_;

  void Shutdown();
  [[nodiscard]] bool InitTls();
  void CreateSocket();
  void CreateTlsSession();
  void Listen();
  std::size_t Read(std::array<uint8_t, 4096> &buffer, bool &more_data);
};

typedef std::shared_ptr<TcpClient> PTcpClient;

}

#endif //LIBC1_NET_SRC_TCPCLIENT_HPP_
