/* Copyright 2022 Sensaru GmbH */

#ifndef LIBC1_NET_SRC_TCPSERVER_H_
#define LIBC1_NET_SRC_TCPSERVER_H_

#include "TcpCommon.h"
#include "TcpSocket.h"
#include "TlsCommon.h"
#include "IQueue.h"
#include "ProxyProtocol.h"

#include <memory>
#include <functional>
#include <queue>
#include <map>
#include <unordered_map>

namespace C1Net {

/**
 * Class to easily create a TCP server.
 *
 * TCP Server Example Code
 * =======================
 *
 * This example creates a simple TCP server that listens on port 8082. The server echos all bytes received. You can
 * e. g. connect to the server using
 *
 *     telnet localhost 8082
 *
 * Save the example below in `main.cpp` and compile with:
 *
 *     g++ -o main -std=c++17 main.cpp -lc1-net -lgcrypt -lgnutls
 *
 * Start the example:
 *
 *     ./main
 *
 * Example code:
 *
 *    #include <c1-net/TcpServer.h>
 *
 *    #include <iostream>
 *
 *    std::unique_ptr<C1Net::TcpServer> tcp_server_;
 *
 *    void Log(uint32_t log_level, const std::string &message) {
 *      std::cout << "Log message (" << log_level << "): " << message << std::endl;
 *    }
 *
 *    void NewConnection(const C1Net::TcpServer::PTcpClientData &client_data) {
 *      std::cout << "New connection from " << client_data->GetIpAddress() << " on port " << client_data->GetPort() << ". Client ID: " << client_data->GetId() << std::endl;
 *    }
 *
 *    void PacketReceived(const C1Net::TcpServer::PTcpClientData &client_data, const C1Net::TcpPacket &packet) {
 *      std::string data(packet.begin(), packet.end());
 *      BaseLib::HelperFunctions::trim(data);
 *      std::cout << "Packet received from client " << client_data->GetId() << ": " << data << std::endl;
 *      std::vector<uint8_t> response;
 *      response.push_back('R');
 *      response.push_back(':');
 *      response.push_back(' ');
 *      response.insert(response.end(), packet.begin(), packet.end());
 *      tcp_server_->Send(client_data, response);
 *    }
 *
 *    void ConnectionClosed(const C1Net::TcpServer::PTcpClientData &client_data, int32_t error_code, std::string error_message) {
 *      std::cout << "Connection to client " << client_data->GetId() << " closed (" << error_code << ", " << error_message << ")." << std::endl;
 *    }
 *
 *    int main() { *
 *      C1Net::TcpServer::TcpServerInfo server_info;
 *      server_info.listen_address = "::";
 *      server_info.port = 8082;
 *
 *      server_info.log_callback = std::bind(&Log, std::placeholders::_1, std::placeholders::_2);
 *      server_info.new_connection_callback = std::bind(&NewConnection, std::placeholders::_1);
 *      //For class methods add `this` as first parameter:
 *      //server_info.newConnectionCallback = std::bind(&MyClass::NewConnection, this, std::placeholders::_1);
 *      server_info.packet_received_callback = std::bind(&PacketReceived, std::placeholders::_1, std::placeholders::_2);
 *      server_info.connection_closed_callback = std::bind(&ConnectionClosed, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3);
 *
 *      tcp_server_ = std::make_unique<C1Net::TcpServer>(server_info);
 *
 *      tcp_server_->Start();
 *      std::cout << "Started listening." << std::endl;
 *
 *      //Alternatively you can call `Bind()` as root first, then drop privileges and start the server:
 *      //tcp_server_->Bind();
 *      //...
 *      //Drop privileges
 *      //...
 *      //_tcpServer->Start();
 *
 *      for (int32_t i = 0; i < 300; i++) //Run for 300 seconds
 *      {
 *        std::this_thread::sleep_for(std::chrono::milliseconds(1000));
 *      }
 *
 *      tcp_server_->Stop();
 *      tcp_server_->WaitForServerStopped();
 *    }
 */
class TcpServer : public IQueue {
 public:
  struct AverageMeanData {
    std::atomic<int64_t> last_measurement{0};
    std::atomic<double> last_output{0.0};
  };

  struct ClientStatistics {
    int32_t client_id = -1;
    int32_t thread_index = -1;
    double average_packets_per_minute_received = 0.0;
    double average_packets_per_minute_sent = 0.0;
  };
  typedef std::shared_ptr<ClientStatistics> PClientStatistics;

  class TcpClientData {
   public:
    friend class TcpServer;

    int32_t GetId() const;
    std::string GetIpAddress();
    uint16_t GetPort();
    const std::string &GetClientCertDn() const;
    const std::string &GetClientCertSerial() const;
    int64_t GetClientCertExpiration() const;

    void Shutdown();
   private:
    int32_t id = 0;

    std::mutex properties_mutex;

    std::string ip_address;
    uint16_t port = 0;

    int32_t thread_index = -1;
    int epoll_fd = -1;
    PTcpSocket socket;
    int64_t last_incoming_data = 0;
    std::array<uint8_t, 4096> buffer{};
    std::unordered_map<std::string, std::shared_ptr<CertificateCredentials>> certificate_credentials;
    std::string client_cert_dn;
    std::string client_cert_serial;
    int64_t client_cert_expiration = -1;
    /**
     * Mutex for `busy` and `backlog`
     */
    std::mutex backlog_and_busy_mutex;
    bool busy = false;
    std::queue<std::shared_ptr<TcpPacket>> backlog;

    AverageMeanData average_packets_per_minute_received;
    AverageMeanData average_packets_per_minute_sent;

    void SetIpAddress(const std::string &value);
    void SetPort(uint16_t value);
  };
  typedef std::shared_ptr<TcpClientData> PTcpClientData;

  struct TcpServerInfo {
    std::string listen_address;

    /**
     * If set to 0, a random port is selected.
     */
    uint16_t port = 0;

    uint32_t connection_backlog_size = 100;
    uint32_t max_connections = 100;

    /**
     * Socket read timeout in milliseconds.
     */
    uint64_t socket_read_timeout = 15000;

    /**
     * Socket write timeout in milliseconds.
     */
    uint64_t socket_write_timeout = 15000;

    /**
     * Data must arrive at least once per this interval (in milliseconds).
     */
    uint64_t min_incoming_data_interval = 3600000;

    bool tls = false;
    std::unordered_map<std::string, PCertificateInfo> certificates;
    bool require_client_cert = false;

    /**
     * Use proxy protocol to read original sender information. Please note that non-proxy protocol packets are
     * discarded when this is enabled.
     *
     * See https://www.haproxy.org/download/1.8/doc/proxy-protocol.txt
     */
    bool use_proxy_protocol = false;

    /**
     * Only relevant when `use_proxy_protocol` is set to `true`. List of allowed IPv4 and IPv6 source addresses. All
     * other source addresses are rejected.
     */
    std::vector<std::string> valid_proxy_protocol_sources;

    uint32_t connect_threads = 1;
    uint32_t listen_threads = 1;
    uint32_t processing_threads = 1;

    std::function<void(uint32_t log_level, const std::string &message)> log_callback;
    std::function<void(const PTcpClientData &client_data)> new_connection_callback;
    std::function<void(const PTcpClientData &client_data, int32_t error_code, const std::string &error_string)> connection_closed_callback;

    /**
     * Called for every received packet. Reading the client does not continue until the callback method returns. So make
     * sure, processing is fast.
     */
    std::function<void(const PTcpClientData &client_data, const TcpPacket &packet)> packet_received_callback;
  };

  explicit TcpServer(const TcpServerInfo &tcp_server_info);
  ~TcpServer() override;

  /**
   * Checks if one of the listen threads is hanging. For every hanging thread an output is send to log_callback with
   * log level 2.
   *
   * @return Returns `true` when everything is ok and `false` if at least one thread is hanging.
   */
  bool Lifetick();
  std::size_t GetProcessingQueueSize();
  std::size_t GetClientCount();

  double GetPacketsPerMinuteReceived();
  double GetPacketsPerMinuteSent();

  std::vector<PClientStatistics> GetClientStatistics();

  /**
   * Returns the current server thread load. A load of 1.0 means all threads are in use. (Queue backlog / thread count) is added to this load, so the load can be greater than 1.0.
   *
   * @return
   */
  double GetServerThreadLoad();

  /**
   * Returns the current processing thread load. A load of 1.0 means all threads are in use. (Queue backlog / thread count) is added to this load, so the load can be greater than 1.0.
   *
   * @return
   */
  double GetProcessingThreadLoad();

  /**
   * Returns the maximum load of the last minute.
   *
   * @return
   */
  double GetProcessingThreadLoadMax();

  void SetCertificates(const std::unordered_map<std::string, PCertificateInfo> &certificates);

  /**
   * Get the client data by client ID.
   *
   * @param client_id The client ID to return the client data for.
   * @return Returns the client data. If the client ID is not found, a nullptr is returned.
   */
  PTcpClientData GetClientData(int client_id);

  /**
   * Binds a socket to an IP address and a port. This splits up the start Process to be able to listen on ports
   * lower than 1024 and do a privilege drop. Call Start() to start listening.
   *
   * @see Start
   *
   * @param[out] listen_address The IP address the server was bound to (e. g. `192.168.0.152`).
   */
  void Bind();
  void Start();
  void Stop();
  void WaitForServerStopped();

  void ReloadCertificates();

  bool Send(const PTcpClientData &client_data, const uint8_t *packet, std::size_t packet_size, bool close_connection = false);
  bool Send(const PTcpClientData &client_data, const TcpPacket &packet, bool close_connection = false);
  bool Send(const PTcpClientData &client_data, const std::vector<char> &packet, bool close_connection = false);
  bool Send(int client_id, const uint8_t *packet, std::size_t packet_size, bool close_connection = false);
  bool Send(int client_id, const TcpPacket &packet, bool close_connection = false);
  bool Send(int client_id, const std::vector<char> &packet, bool close_connection = false);

  /**
   * Sends a packet to all connected clients.
   * @param packet
   */
  void Broadcast(const TcpPacket &packet);
 private:
  class QueueEntry : public IQueueEntry {
   public:
    QueueEntry() = default;

    explicit QueueEntry(const PTcpClientData &client_data) {
      this->client_data = client_data;
    }

    PTcpClientData client_data;
  };

  class ConnectQueueEntry : public QueueEntry {
   public:
    ConnectQueueEntry() = default;

    explicit ConnectQueueEntry(const PTcpClientData &client_data, const PSocket &socket) : QueueEntry(client_data) {
      this->client_data = client_data;
      this->socket = socket;
    }

    PTcpClientData client_data;
    PSocket socket;
  };

  TcpServerInfo tcp_server_info_;
  std::vector<std::array<uint8_t, 16>> valid_proxy_protocol_sources_;

  std::atomic_bool socket_bound_{false};
  std::mutex socket_mutex_;
  PSocket socket_;

  std::atomic_bool stop_server_{false};
  std::vector<std::thread> listen_threads_;

  std::mutex clients_mutex_;
  std::map<int32_t, PTcpClientData> clients_;

  //{{{ TLS
  bool tls_init_failed_ = false;
  PTlsPriorityCache tls_priority_cache_;
  std::mutex certificate_credentials_mutex_;

  /**
   * Stores the certificate credentials so that they can be replaced at any time.
   */
  std::unordered_map<std::string, std::shared_ptr<CertificateCredentials>> certificate_credentials_;
  //}}}

  //{{{ Monitoring/statistics
  std::atomic<uint32_t> server_threads_in_use_{0};
  std::deque<std::atomic<int64_t>> listen_thread_life_ticks_;
  std::deque<AverageMeanData> average_packets_per_minute_received_;
  std::deque<AverageMeanData> average_packets_per_minute_sent_;
  //}}}

  static int TlsPostClientHello(gnutls_session_t tls_session);
  void CollectGarbage();
  void CollectGarbage(std::unordered_map<int32_t, PTcpClientData> &clients);
  void ConnectClient(TcpServer::PTcpClientData &client_data, PSocket &socket);
  bool InitTls();
  bool InitTlsPriorityCache();
  void InitClientTls(const TcpServer::PTcpClientData &client_data, PSocket &client_socket, PTlsSession &tls_session);
  void BindSocket();
  void Listen(int32_t thread_index);
  void ReadClient(PTcpClientData &client_data);
  static void ShutdownClientSocket(const PTcpClientData &client_data);
  void ProcessQueueEntry(int32_t index, std::shared_ptr<IQueueEntry> &entry) override;
  void QueueLog(uint32_t log_level, const std::string &message) override;
};

typedef std::shared_ptr<TcpServer> PTcpServer;

}

#endif //LIBC1_NET_SRC_TCPSERVER_H_
