/* Copyright 2022 Sensaru GmbH */

#ifndef LIBC1_NET_SRC_TCPSOCKET_H_
#define LIBC1_NET_SRC_TCPSOCKET_H_

#include "Socket.h"
#include "TlsCommon.h"
#include "TcpCommon.h"

#include <vector>
#include <functional>
#include <mutex>
#include <atomic>
#include <thread>

namespace C1Net {

struct TcpSocketInfo {
  uint32_t read_timeout = 15000;
  uint32_t write_timeout = 15000;

  std::function<void(uint32_t log_level, const std::string &message)> log_callback;
};

struct TcpSocketHostInfo {
  std::string host;
  uint16_t port = 0;
  bool tls = false;
  bool verify_certificate = true;
  std::string ca_file;
  std::string ca_data;
  std::string client_cert_file;
  std::string client_cert_data;
  std::string client_key_file;
  std::string client_key_data;
  bool verify_custom_hostname = false;
  std::string custom_hostname;

  bool auto_connect = true;
  uint32_t connection_retries = 3;
};

class TcpSocket {
 public:
  TcpSocket(const TcpSocketInfo &tcp_socket_info, const TcpSocketHostInfo &tcp_socket_host_info);
  TcpSocket(const TcpSocketInfo &tcp_socket_info, PSocket &socket, PTlsSession tls_session = nullptr);
  ~TcpSocket();

  int GetSocketHandle();
  gnutls_session_t GetTlsSessionHandle();
  std::string GetIpAddress();

  void SetHostname(const std::string &hostname);
  void SetPort(uint16_t port);
  void SetReadTimeout(uint32_t value);
  void SetWriteTimeout(uint32_t value);

  void SetVerifyHostname(bool value);

  bool Connected();
  bool IsValid();

  void Open();
  void Shutdown();

  /**
   * Reads data from socket into a buffer. Make sure, this method is called from one thread at a time.
   *
   * @param buffer
   * @param buffer_size
   * @param more_data Only used for TLS connections. Set to true when there is a pending record.
   * @return
   */
  std::size_t Read(uint8_t *buffer, std::size_t buffer_size, bool &more_data);

  /**
   * Reads data from socket into a buffer without removing the read data. Make sure, this method is called from one
   * thread at a time. This only works for connections NOT using TLS. Also the connection must be established before
   * this method is called.
   *
   * @param buffer
   * @param buffer_size
   * @return
   */
  std::size_t Peek(uint8_t *buffer, std::size_t buffer_size);
  void Send(const TcpPacket &packet);
  void Send(const uint8_t *buffer, std::size_t bytes_to_write);
 private:
  TcpSocketInfo tcp_socket_info_;
  TcpSocketHostInfo tcp_socket_host_info_;

  std::mutex properties_mutex_;

  PSocket socket_ = std::make_shared<Socket>(-1);

  std::string hostname_;
  uint16_t port_ = 0;
  std::atomic<uint32_t> read_timeout_{15000};
  std::atomic<uint32_t> write_timeout_{15000};
  std::string ip_address_;

  //{{{ TLS
  bool tls_init_failed_ = false;
  PTlsSession tls_session_;
  std::mutex certificate_credentials_mutex_;

  PCertificateInfo certificates_;

  /**
   * Contains a copy of the current credentials in case the credentials are replaced while the connection is open.
   */
  std::shared_ptr<CertificateCredentials> current_client_certificate_credentials_;

  /**
   * Stores the certificate credentials so that they can be replaced at any time.
   */
  std::shared_ptr<CertificateCredentials> certificate_credentials_;
  //}}}

  std::atomic_bool connecting_{false};

  [[nodiscard]] bool InitTls();
  void CreateSocket();
  void CreateTlsSession();
};

typedef std::shared_ptr<TcpSocket> PTcpSocket;

}

#endif //LIBC1_NET_SRC_TCPSOCKET_H_
