/* Copyright 2022 Sensaru GmbH */

#ifndef LIBC1_NET_SRC_UDPCLIENT_H_
#define LIBC1_NET_SRC_UDPCLIENT_H_

#include "Socket.h"
#include "Exceptions.h"

#include <vector>
#include <functional>
#include <mutex>
#include <atomic>
#include <thread>

namespace C1Net {

typedef std::vector<uint8_t> UdpPacket;

/**
 * Class to easily create a UDP client.
 *
 * UDP Client Example Code
 * =======================
 *
 * This example creates a simple UDP client that connects to 127.0.0.1:9998. The client sends "Hi\n" every second and
 * prints all incoming packets to standard output.
 *
 * Save the example below in `main.cpp` and compile with:
 *
 *     g++ -o main -std=c++17 main.cpp -lc1-net -lgcrypt -lgnutls
 *
 * Start and connect using e. g. netcat:
 *
 *     nc -l -s :: -p 9998 -u
 *     ./main
 *
 * Example code:
 *
 *     #include <c1-net/UdpClient.h>
 *
 *     #include <iostream>
 *
 *     void Log(uint32_t log_level, const std::string &message) {
 *         std::cout << message << std::endl;
 *     }
 *
 *     void PacketReceived(const C1Net::UdpPacket &packet) {
 *         std::cout << BaseLib::HelperFunctions::getHexString(packet) << std::endl;
 *     }
 *
 *     int main() { *
 *         C1Net::UdpClientInfo udp_client_info;
 *         udp_client_info.host = "127.0.0.1";
 *         udp_client_info.port = 9998;
 *         udp_client_info.log_callback = std::bind(&Log, std::placeholders::_1, std::placeholders::_2);
 *         udp_client_info.packet_received_callback = std::bind(&PacketReceived, std::placeholders::_1);
 *
 *         C1Net::UdpClient udp_client(udp_client_info);
 *         udp_client.Start();
 *
 *         C1Net::UdpPacket packet = {0x48, 0x69, 0x0A};
 *
 *         while (true) {
 *             std::this_thread::sleep_for(std::chrono::seconds(1));
 *             udp_client.Send(packet);
 *         }
 *     }
 */
class UdpClient {
 public:
  struct UdpClientInfo {
    std::string host;
    uint16_t port = 0;

    std::function<void(uint32_t log_level, const std::string &message)> log_callback;
    std::function<void(const UdpPacket &packet)> packet_received_callback;
  };

  explicit UdpClient(const UdpClientInfo &udp_client_info);
  ~UdpClient();

  void Start();
  void Stop();
  void WaitForClientStopped();

  void Send(const UdpPacket &packet);
 private:
  UdpClientInfo udp_client_info_;

  std::mutex socket_mutex_;
  PSocket socket_ = std::make_shared<Socket>(-1);

  std::atomic_bool stop_client_{false};
  std::thread listen_thread_;

  void CreateSocket();
  void Listen();
};

}

#endif //LIBC1_NET_SRC_UDPCLIENT_H_
