/* Copyright 2022 Sensaru GmbH */

#ifndef LIBC1_NET_SRC_UDPSERVER_H_
#define LIBC1_NET_SRC_UDPSERVER_H_

#include "Socket.h"
#include "Exceptions.h"
#include "IQueue.h"

#include <vector>
#include <functional>
#include <mutex>
#include <atomic>
#include <thread>

namespace C1Net {

typedef std::vector<uint8_t> UdpPacket;

/**
 * Class to easily create a UDP server.
 *
 * UDP Server Example Code
 * =======================
 *
 * This example creates a simple UDP server that binds to 0.0.0.0:9999. The server echos back all received packets.
 *
 * Save the example below in `main.cpp` and compile with:
 *
 *     g++ -o main -std=c++17 main.cpp -lc1-net -lgcrypt -lgnutls
 *
 * Start and connect using e. g. netcat:
 *
 *     ./main
 *     nc -u 127.0.0.1 9999
 *
 * Example code:
 *
 *     #include <c1-net/UdpServer.h>
 *
 *     #include <iostream>
 *
 *     std::unique_ptr<C1Net::UdpServer> server;
 *
 *     void Log(uint32_t log_level, const std::string &message) {
 *         std::cout << message << std::endl;
 *     }
 *
 *     void PacketReceived(const C1Net::PUdpClientData &client_data, const C1Net::UdpPacket &packet) {
 *         std::cout << client_data->ip_address << " " << client_data->port << " " << BaseLib::HelperFunctions::getHexString(packet) << std::endl;
 *
 *         server->Send(client_data, packet);
 *     }
 *
 *     int main() {
 *         C1Net::UdpServerInfo server_info;
 *
 *         server_info.listen_address = "0.0.0.0";
 *         server_info.port = 9999;
 *
 *         server_info.log_callback = std::bind(&Log, std::placeholders::_1, std::placeholders::_2);
 *         server_info.packet_received_callback = std::bind(&PacketReceived, std::placeholders::_1, std::placeholders::_2);
 *
 *         server = std::make_unique<C1Net::UdpServer>(server_info);
 *         server->Start();
 *
 *         while (true) {
 *                 std::this_thread::sleep_for(std::chrono::seconds(1));
 *         }
 *     }
 */
class UdpServer : public IQueue {
 public:
  struct UdpClientData {
    std::string ip_address;
    uint16_t port = 0;
  };
  typedef std::shared_ptr<UdpClientData> PUdpClientData;

  struct UdpServerInfo {
    std::string listen_address;
    uint16_t port = 0;
    uint32_t listen_threads = 1;
    uint32_t processing_threads = 1;

    std::function<void(uint32_t log_level, const std::string &message)> log_callback;
    std::function<void(const PUdpClientData &client_data, const UdpPacket &packet)> packet_received_callback;
  };

  explicit UdpServer(const UdpServerInfo &udp_server_info);
  ~UdpServer() override;

  void Start();
  void Stop();
  void WaitForServerStopped();

  void Send(const PUdpClientData &client_data, const UdpPacket &packet);
 private:
  class QueueEntry : public IQueueEntry {
   public:
    QueueEntry() = default;

    explicit QueueEntry(const PUdpClientData &client_data, const UdpPacket &packet) {
      this->client_data = client_data;
      this->packet = packet;
    }

    ~QueueEntry() override = default;

    PUdpClientData client_data;
    UdpPacket packet;
  };

  UdpServerInfo udp_server_info_;

  std::mutex socket_mutex_;
  PSocket socket_;

  std::atomic_bool stop_server_{false};
  std::vector<std::thread> listen_threads_;

  void BindSocket();
  void Listen();
  void ProcessQueueEntry(int32_t index, std::shared_ptr<IQueueEntry> &entry) override;
  void QueueLog(uint32_t log_level, const std::string &message);
};

}

#endif //LIBC1_NET_SRC_UDPSERVER_H_
